from enum import Enum

from . import common
from . import equities
from . import generic
from . import options

import sys
assert sys.version_info[0] >= 3

__error_message = (
        'EquityOrderBuilder has been deleted from the library. Please use ' +
        'OrderBuilder and its associated templates instead. See here for ' +
        'details: https://tda-api.readthedocs.io/en/latest/' +
        'order-templates.html#what-happened-to-equityorderbuilder')

if sys.version_info[1] >= 7:
    def __getattr__(name):
        if name == 'EquityOrderBuilder':
            raise ImportError(__error_message)
        raise AttributeError(name)
else:  # pragma: no cover
    class EquityOrderBuilder:
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(globals()['__error_message'])



'''Implements additional functionality beyond what's implemented in the client
module.'''

import datetime
import dateutil.parser
import enum
import httpx
import inspect
import re


def class_fullname(o):
    return o.__module__ + '.' + o.__name__


class EnumEnforcer:
    def __init__(self, enforce_enums):
        self.enforce_enums = enforce_enums

    def type_error(self, value, required_enum_type):
        possible_members_message = ''

        if isinstance(value, str):
            possible_members = []
            for member in required_enum_type.__members__:
                fullname = class_fullname(required_enum_type) + '.' + member
                if value in fullname:
                    possible_members.append(fullname)

            # Oxford comma insertion
            if possible_members:
                possible_members_message = 'Did you mean ' + ', '.join(
                    possible_members[:-2] + [' or '.join(
                        possible_members[-2:])]) + '? '

        raise ValueError(
            ('expected type "{}", got type "{}". {}(initialize with ' +
             'enforce_enums=False to disable this checking)').format(
                required_enum_type.__name__,
                type(value).__name__,
                possible_members_message))

    def convert_enum(self, value, required_enum_type):
        if value is None:
            return None

        if isinstance(value, required_enum_type):
            return value.value
        elif self.enforce_enums:
            self.type_error(value, required_enum_type)
        else:
            return value

    def convert_enum_iterable(self, iterable, required_enum_type):
        if iterable is None:
            return None

        if isinstance(iterable, required_enum_type):
            return [iterable.value]

        values = []
        for value in iterable:
            if isinstance(value, required_enum_type):
                values.append(value.value)
            elif self.enforce_enums:
                self.type_error(value, required_enum_type)
            else:
                values.append(value)
        return values

    def set_enforce_enums(self, enforce_enums):
        self.enforce_enums = enforce_enums


class UnsuccessfulOrderException(ValueError):
    '''
    Raised by :meth:`Utils.extract_order_id` when attempting to extract an
    order ID from a :meth:`Client.place_order` response that was not successful.
    '''


class AccountIdMismatchException(ValueError):
    '''
    Raised by :meth:`Utils.extract_order_id` when attempting to extract an
    order ID from a :meth:`Client.place_order` with a different account ID than
    the one with which the :class:`Utils` was initialized.
    '''


class LazyLog:
    'Helper to defer evaluation of expensive variables in log messages'
    def __init__(self, func):
        self.func = func
    def __str__(self):
        return self.func()


class Utils(EnumEnforcer):
    '''Helper for placing orders on equities. Provides easy-to-use
    implementations for common tasks such as market and limit orders.'''

    def __init__(self, client, account_hash):
        '''Creates a new ``Utils`` instance. For convenience, this object
        assumes the user wants to work with a single account ID at a time.'''
        super().__init__(True)

        self.client = client
        self.account_hash = account_hash

    def set_account_hash(self, account_hash):
        '''Set the account ID used by this ``Utils`` instance.'''
        self.account_hash = account_hash

    def extract_order_id(self, place_order_response):
        '''Attempts to extract the order ID from a response object returned by
        :meth:`Client.place_order() <schwab.client.Client.place_order>`. Return
        ``None`` if the order location is not contained in the response.

        :param place_order_response: Order response as returned by
                                     :meth:`Client.place_order()
                                     <schwab.client.Client.place_order>`. Note this
                                     method requires that the order was
                                     successful.

        :raise ValueError: if the order was not succesful or if the order's
                           account ID is not equal to the account ID set in this
                           ``Utils`` object.
        '''
        if place_order_response.is_error:
            raise UnsuccessfulOrderException(
                'order not successful: status {}'.format(place_order_response.status_code))

        try:
            location = place_order_response.headers['Location']
        except KeyError:
            return None

        m = re.match(
                r'https://api.schwabapi.com/trader/v1/accounts/([/w]+)/orders/(/d+)',
                location)

        if m is None:
            return None
        account_hash, order_id = m.group(1), int(m.group(2))

        if str(account_hash) != str(self.account_hash):
            raise AccountIdMismatchException(
                'order request account hash != Utils.account_hash')

        return order_id



##########################################################################
# Authentication Wrappers

from authlib.integrations.httpx_client import AsyncOAuth2Client, OAuth2Client
from prompt_toolkit import prompt

import json
import logging
import os
import pickle
import sys
import time
import warnings

from schwab.client import AsyncClient, Client
from schwab.debug import register_redactions


TOKEN_ENDPOINT = 'https://api.schwabapi.com/v1/oauth/token'


def get_logger():
    return logging.getLogger(__name__)


def __update_token(token_path):
    def update_token(t, *args, **kwargs):
        get_logger().info('Updating token to file %s', token_path)

        with open(token_path, 'w') as f:
            json.dump(t, f)
    return update_token


def __token_loader(token_path):
    def load_token():
        get_logger().info('Loading token from file %s', token_path)

        with open(token_path, 'rb') as f:
            token_data = f.read()
            try:
                return json.loads(token_data.decode())
            except ValueError:
                get_logger().warning(
                    'Unable to load JSON token from file %s, ' +
                    'falling back to pickle', token_path)
                return pickle.loads(token_data)
    return load_token


def client_from_token_file(token_path, api_key, app_secret, asyncio=False,
                           enforce_enums=True):
    '''
    Returns a session from an existing token file. The session will perform
    an auth refresh as needed. It will also update the token on disk whenever
    appropriate.

    :param token_path: Path to an existing token. Updated tokens will be written
                       to this path. If you do not yet have a token, use
                       :func:`~schwab.auth.client_from_login_flow` or
                       :func:`~schwab.auth.easy_client` to create one.
    :param api_key: Your Schwab application's app key.
    :param asyncio: If set to ``True``, this will enable async support allowing
                    the client to be used in an async environment. Defaults to
                    ``False``
    :param enforce_enums: Set it to ``False`` to disable the enum checks on ALL
                          the client methods. Only do it if you know you really
                          need it. For most users, it is advised to use enums
                          to avoid errors.
    '''

    load = __token_loader(token_path)

    return client_from_access_functions(
        api_key, app_secret, load, __update_token(token_path), asyncio=asyncio,
        enforce_enums=enforce_enums)


def __fetch_and_register_token_from_redirect(
        oauth, redirected_url, api_key, app_secret, token_path,
        token_write_func, asyncio, enforce_enums=True):
    token = oauth.fetch_token(
        TOKEN_ENDPOINT,
        authorization_response=redirected_url,
        client_id=api_key, auth=(api_key, app_secret))

    # Don't emit token details in debug logs
    register_redactions(token)

    # Set up token writing and perform the initial token write
    update_token = (
        __update_token(token_path) if token_write_func is None
        else token_write_func)
    metadata_manager = TokenMetadata(int(time.time()), update_token)
    update_token = metadata_manager.wrapped_token_write_func()
    update_token(token)

    # The synchronous and asynchronous versions of the OAuth2Client are similar
    # enough that can mostly be used interchangeably. The one currently known
    # exception is the token update function: the synchronous version expects a
    # synchronous one, the asynchronous requires an async one. The
    # oauth_client_update_token variable will contain the appropriate one.
    if asyncio:
        async def oauth_client_update_token(t, *args, **kwargs):
            update_token(t, *args, **kwargs)  # pragma: no cover
        session_class = AsyncOAuth2Client
        client_class = AsyncClient
    else:
        oauth_client_update_token = update_token
        session_class = OAuth2Client
        client_class = Client

    # Return a new session configured to refresh credentials
    return client_class(
        api_key,
        session_class(api_key,
                      client_secret=app_secret,
                      token=token,
                      update_token=oauth_client_update_token),
        token_metadata=metadata_manager, enforce_enums=enforce_enums)


class RedirectTimeoutError(Exception):
    pass


class TokenMetadata:
    '''
    Provides the functionality required to maintain and update our view of the
    token's metadata.
    '''
    # XXX: Token metadata is currently not considered sensitive enough to wrap
    #      register for redactions. If we add anything sensitive to the token
    #      metadata, we'll need to update the redaction registration logic.

    def __init__(self, creation_timestamp, unwrapped_token_write_func=None):
        self.creation_timestamp = creation_timestamp

        # The token write function is ultimately stored in the session. When we
        # get a new token we immediately wrap it in a new sesssion. We hold on
        # to the unwrapped token writer function to allow us to inject the
        # appropriate write function.
        self.unwrapped_token_write_func = unwrapped_token_write_func

    @classmethod
    def from_loaded_token(cls, token, app_secret, unwrapped_token_write_func=None):
        '''
        Returns a new ``TokenMetadata`` object extracted from the metadata of
        the loaded token object. If the token has a legacy format which contains
        no metadata, assign default values.
        '''
        logger = get_logger()
        if cls.is_metadata_aware_token(token):
            logger.info(
                    'Loaded metadata aware token with creation timestamp %s',
                    token['creation_timestamp'])
            return TokenMetadata(
                token['creation_timestamp'], unwrapped_token_write_func)
        elif cls.is_legacy_token(token):
            logger.info('Loaded legacy token')
            return TokenMetadata(None, unwrapped_token_write_func)
        else:
            logger.warn('Unrecognized token format')
            return TokenMetadata(None, unwrapped_token_write_func)

    @classmethod
    def is_legacy_token(cls, token):
        return 'creation_timestamp' not in token

    @classmethod
    def is_metadata_aware_token(cls, token):
        return 'creation_timestamp' in token and 'token' in token

    def wrapped_token_write_func(self):
        '''
        Hook the call to the token write function so that the write function is
        passed the metadata-aware version of the token.
        '''
        def wrapped_token_write_func(token, *args, **kwargs):
            return self.unwrapped_token_write_func(
                self.wrap_token_in_metadata(token), *args, **kwargs)
        return wrapped_token_write_func

    def wrap_token_in_metadata(self, token):
        return {
            'creation_timestamp': self.creation_timestamp,
            'token': token,
        }


# TODO: Raise an exception when passing both token_path and token_write_func
def client_from_login_flow(webdriver, api_key, app_secret, callback_url, token_path,
                           redirect_wait_time_seconds=0.1, max_waits=3000,
                           asyncio=False, token_write_func=None,
                           enforce_enums=True):
    '''
    Uses the webdriver to perform an OAuth webapp login flow and creates a
    client wrapped around the resulting token. The client will be configured to
    refresh the token as necessary, writing each updated version to
    ``token_path``.

    **Warning:** Schwab appears to block logins performed within a webdriver. 
    This library has been included as a direct copy from ``tda-api``, but it may 
    be removed in the future.

    :param webdriver: `selenium <https://selenium-python.readthedocs.io>`__
                      webdriver which will be used to perform the login flow.
    :param api_key: Your Schwab application's app key.
    :param callback_url: Your Schwab application's callback URL. Note this must
                         *exactly* match the value you've entered in your
                         application configuration, otherwise login will fail
                         with a security error.
    :param token_path: Path to which the new token will be written. If the token
                       file already exists, it will be overwritten with a new
                       one. Updated tokens will be written to this path as well.
    :param asyncio: If set to ``True``, this will enable async support allowing
                    the client to be used in an async environment. Defaults to
                    ``False``
    :param enforce_enums: Set it to ``False`` to disable the enum checks on ALL
                          the client methods. Only do it if you know you really
                          need it. For most users, it is advised to use enums
                          to avoid errors.
    '''
    get_logger().info('Creating new token with redirect URL \'%s\' ' +
                       'and token path \'%s\'', callback_url, token_path)

    oauth = OAuth2Client(api_key, callback_url=callback_url)
    authorization_url, state = oauth.create_authorization_url(
        'https://api.schwabapi.com/v1/oauth/authorize')

    # Open the login page and wait for the redirect
    print('\n**************************************************************\n')
    print('Opening the login page in a webdriver. Please use this window to',
          'log in. Successful login will be detected automatically.')
    print()
    print('If you encounter any issues, see here for troubleshooting: ' +
          'https://schwab-py.readthedocs.io/en/stable/auth.html' +
          '#troubleshooting')
    print('\n**************************************************************\n')

    webdriver.get(authorization_url)

    # Tolerate redirects to HTTPS on the callback URL
    if callback_url.startswith('http://'):
        print(('WARNING: Your callback URL ({}) will transmit data over HTTP, ' +
               'which is a potentially severe security vulnerability. ' +
               'Please go to your app\'s configuration with Schwab ' +
               'and update your callback URL to begin with \'https\' ' +
               'to stop seeing this message.').format(callback_url))

        callback_urls = (callback_url, 'https' + callback_url[4:])
    else:
        callback_urls = (callback_url,)

    # Wait until the current URL starts with the callback URL
    current_url = ''
    num_waits = 0
    while not any(current_url.startswith(r_url) for r_url in callback_urls):
        current_url = webdriver.current_url

        if num_waits > max_waits:
            raise RedirectTimeoutError('timed out waiting for redirect')
        time.sleep(redirect_wait_time_seconds)
        num_waits += 1

    return __fetch_and_register_token_from_redirect(
        oauth, current_url, api_key, app_secret, token_path, token_write_func,
        asyncio, enforce_enums=enforce_enums)


def client_from_manual_flow(api_key, app_secret, callback_url, token_path,
                            asyncio=False, token_write_func=None,
                            enforce_enums=True):
    '''
    Walks the user through performing an OAuth login flow by manually
    copy-pasting URLs, and returns a client wrapped around the resulting token.
    The client will be configured to refresh the token as necessary, writing
    each updated version to ``token_path``.

    Note this method is more complicated and error prone, and should be avoided
    in favor of :func:`client_from_login_flow` wherever possible.

    :param api_key: Your Schwab application's app key.
    :param callback_url: Your Schwab application's callback URL. Note this must
                         *exactly* match the value you've entered in your
                         application configuration, otherwise login will fail
                         with a security error.
    :param token_path: Path to which the new token will be written. If the token
                       file already exists, it will be overwritten with a new
                       one. Updated tokens will be written to this path as well.
    :param asyncio: If set to ``True``, this will enable async support allowing
                    the client to be used in an async environment. Defaults to
                    ``False``
    :param enforce_enums: Set it to ``False`` to disable the enum checks on ALL
                          the client methods. Only do it if you know you really
                          need it. For most users, it is advised to use enums
                          to avoid errors.
    '''
    get_logger().info('Creating new token with callback URL \'%s\' ' +
                       'and token path \'%s\'', callback_url, token_path)

    oauth = OAuth2Client(api_key, redirect_uri=callback_url)
    authorization_url, state = oauth.create_authorization_url(
        'https://api.schwabapi.com/v1/oauth/authorize')

    print('\n**************************************************************\n')
    print('This is the manual login and token creation flow for schwab-py.')
    print('Please follow these instructions exactly:')
    print()
    print(' 1. Open the following link by copy-pasting it into the browser')
    print('    of your choice:')
    print()
    print('        ' + authorization_url)
    print()
    print(' 2. Log in with your account credentials. You may be asked to')
    print('    perform two-factor authentication using text messaging or')
    print('    another method, as well as whether to trust the browser.')
    print()
    print(' 3. When asked whether to allow your app access to your account,')
    print('    select "Allow".')
    print()
    print(' 4. Your browser should be redirected to your callback URI. Copy')
    print('    the ENTIRE address, paste it into the following prompt, and press')
    print('    Enter/Return.')
    print()
    print('If you encounter any issues, see here for troubleshooting:')
    print('https://schwab-py.readthedocs.io/en/latest/auth.html#troubleshooting')
    print('\n**************************************************************\n')

    if callback_url.startswith('http://'):
        print(('WARNING: Your callback URL ({}) will transmit data over HTTP, ' +
               'which is a potentially severe security vulnerability. ' +
               'Please go to your app\'s configuration with TDAmeritrade ' +
               'and update your callback URL to begin with \'https\' ' +
               'to stop seeing this message.').format(callback_url))

    redirected_url = prompt('Redirect URL> ').strip()

    return __fetch_and_register_token_from_redirect(
        oauth, redirected_url, api_key, app_secret, token_path, token_write_func,
        asyncio, enforce_enums=enforce_enums)


def easy_client(api_key, app_secret, callback_url, token_path,
                webdriver_func=None, asyncio=False, enforce_enums=True):
    '''Convenient wrapper around :func:`client_from_login_flow` and
    :func:`client_from_token_file`. If ``token_path`` exists, loads the token
    from it. Otherwise open a login flow to fetch a new token. Returns a client
    configured to refresh the token to ``token_path``.

    **Warning:** Schwab appears to block logins performed within a webdriver. 
    This library has been included as a direct copy from ``tda-api``, but it may 
    be removed in the future.

    *Reminder:* You should never create the token file yourself or modify it in
    any way. If ``token_path`` refers to an existing file, this method will
    assume that file is valid token and will attempt to parse it.

    :param api_key: Your Schwab application's app key.
    :param callback_url: Your Schwab application's redirect URL. Note this must
                         *exactly* match the value you've entered in your
                         application configuration, otherwise login will fail
                         with a security error.
    :param token_path: Path that new token will be read from and written to. If
                       If this file exists, this method will assume it's valid
                       and will attempt to parse it as a token. If it does not,
                       this method will create a new one using
                       :func:`~schwab.auth.client_from_login_flow`. Updated tokens
                       will be written to this path as well.
    :param webdriver_func: Function that returns a webdriver for use in fetching
                           a new token. Will only be called if the token file
                           cannot be found.
    :param asyncio: If set to ``True``, this will enable async support allowing
                    the client to be used in an async environment. Defaults to
                    ``False``
    :param enforce_enums: Set it to ``False`` to disable the enum checks on ALL
                          the client methods. Only do it if you know you really
                          need it. For most users, it is advised to use enums
                          to avoid errors.
    '''
    logger = get_logger()

    if os.path.isfile(token_path):
        c = client_from_token_file(token_path, api_key, app_secret,
                                   asyncio=asyncio, enforce_enums=enforce_enums)
        logger.info(
                'Returning client loaded from token file \'%s\'', token_path)
        return c
    else:
        logger.warning('Failed to find token file \'%s\'', token_path)

        if webdriver_func is not None:
            with webdriver_func() as driver:
                c = client_from_login_flow(
                    driver, api_key, app_secret, callback_url, token_path,
                    asyncio=asyncio, enforce_enums=enforce_enums)
                logger.info(
                    'Returning client fetched using webdriver, writing' +
                    'token to \'%s\'', token_path)
                return c
        else:
            logger.error('No webdriver_func set, cannot fetch token')
            sys.exit(1)


def client_from_access_functions(api_key, app_secret, token_read_func,
                                 token_write_func, asyncio=False,
                                 enforce_enums=True):
    '''
    Returns a session from an existing token file, using the accessor methods to
    read and write the token. This is an advanced method for users who do not
    have access to a standard writable filesystem, such as users of AWS Lambda
    and other serverless products who must persist token updates on
    non-filesystem places, such as S3. 99.9% of users should not use this
    function.

    Users are free to customize how they represent the token file. In theory,
    since they have direct access to the token, they can get creative about how
    they store it and fetch it. In practice, it is *highly* recommended to
    simply accept the token object and use ``pickle`` to serialize and
    deserialize it, without inspecting it in any way.

    Note the read and write methods must take particular arguments. Please see 
    `this example <https://github.com/alexgolec/schwab-py/tree/master/examples/
    client_from_access_functions.py>`__ for details.

    :param api_key: Your Schwab application's app key.
    :param token_read_func: Function that takes no arguments and returns a token
                            object.
    :param token_write_func: Function that writes the token on update. Will be
                             called whenever the token is updated, such as when
                             it is refreshed. See the above-mentioned example 
                             for what parameters this method takes.
    :param asyncio: If set to ``True``, this will enable async support allowing
                    the client to be used in an async environment. Defaults to
                    ``False``
    :param enforce_enums: Set it to ``False`` to disable the enum checks on ALL
                          the client methods. Only do it if you know you really
                          need it. For most users, it is advised to use enums
                          to avoid errors.
    '''
    token = token_read_func()

    # Extract metadata and unpack the token, if necessary
    metadata = TokenMetadata.from_loaded_token(
            token, app_secret, token_write_func)
    if TokenMetadata.is_metadata_aware_token(token):
        token = token['token']

    # Don't emit token details in debug logs
    register_redactions(token)

    # Return a new session configured to refresh credentials
    #api_key = _normalize_api_key(api_key)

    wrapped_token_write_func = metadata.wrapped_token_write_func()

    if asyncio:
        async def oauth_client_update_token(t, *args, **kwargs):
            wrapped_token_write_func(t, *args, **kwargs)  # pragma: no cover
        session_class = AsyncOAuth2Client
        client_class = AsyncClient
    else:
        oauth_client_update_token = wrapped_token_write_func
        session_class = OAuth2Client
        client_class = Client

    return client_class(
        api_key,
        session_class(api_key,
                      client_secret=app_secret,
                      token=token,
                      token_endpoint=TOKEN_ENDPOINT,
                      update_token=oauth_client_update_token),
        token_metadata=metadata,
        enforce_enums=enforce_enums)



from enum import Enum


class __BaseInstrument:
    def __init__(self, asset_type, symbol):
        self._assetType = asset_type
        self._symbol = symbol


class EquityInstrument(__BaseInstrument):
    '''Represents an equity when creating order legs.'''

    def __init__(self, symbol):
        super().__init__('EQUITY', symbol)


class OptionInstrument(__BaseInstrument):
    '''Represents an option when creating order legs.'''

    def __init__(self, symbol):
        super().__init__('OPTION', symbol)


class InvalidOrderException(Exception):
    '''Raised when attempting to build an incomplete order'''
    pass


class Duration(Enum):
    '''
    Length of time over which the trade will be active.
    '''
    #: Cancel the trade at the end of the trading day. Note if the order cannot
    #: be filled all at once, you may see partial executions throughout the day.
    DAY = 'DAY'

    # TODO: Implement order cancellation date

    #: Keep the trade open for six months, or until the end of the cancel date,
    #: whichever is shorter. Note if the order cannot be filled all at once, you
    #: may see partial executions over the lifetime of the order.
    GOOD_TILL_CANCEL = 'GOOD_TILL_CANCEL'

    #: Either execute the order immediately at the specified price, or cancel it
    #: immediately.
    FILL_OR_KILL = 'FILL_OR_KILL'

    # TODO: Document these
    IMMEDIATE_OR_CANCEL = 'IMMEDIATE_OR_CANCEL'
    END_OF_WEEK = 'END_OF_WEEK'
    END_OF_MONTH = 'END_OF_MONTH'
    NEXT_END_OF_MONTH = 'NEXT_END_OF_MONTH'


class Session(Enum):
    '''
    The market session during which the order trade should be executed.
    '''

    #: Normal market hours, from 9:30am to 4:00pm Eastern.
    NORMAL = 'NORMAL'

    #: Premarket session, from 8:00am to 9:30am Eastern.
    AM = 'AM'

    #: After-market session, from 4:00pm to 8:00pm Eastern.
    PM = 'PM'

    #: Orders are active during all trading sessions except the overnight
    #: session. This is the union of ``NORMAL``, ``AM``, and ``PM``.
    SEAMLESS = 'SEAMLESS'


class OrderType(Enum):
    '''
    Type of order to place.
    '''

    #: Execute the order immediately at the best-available price.
    #: `More Info <https://www.investopedia.com/terms/m/marketorder.asp>`__.
    MARKET = 'MARKET'

    #: Execute the order at your price or better.
    #: `More info <https://www.investopedia.com/terms/l/limitorder.asp>`__.
    LIMIT = 'LIMIT'

    #: Wait until the price reaches the stop price, and then immediately place a
    #: market order.
    #: `More Info <https://www.investopedia.com/terms/l/limitorder.asp>`__.
    STOP = 'STOP'

    #: Wait until the price reaches the stop price, and then immediately place a
    #: limit order at the specified price. `More Info
    #: <https://www.investopedia.com/terms/s/stop-limitorder.asp>`__.
    STOP_LIMIT = 'STOP_LIMIT'

    #: Similar to ``STOP``, except if the price moves in your favor, the stop
    #: price is adjusted in that direction. Places a market order if the stop
    #: condition is met.
    #: `More info <https://www.investopedia.com/terms/t/trailingstop.asp>`__.
    TRAILING_STOP = 'TRAILING_STOP'

    CABINET = 'CABINET'

    NON_MARKETABLE = 'NON_MARKETABLE'

    #: Place the order at the closing price immediately upon market close.
    #: `More info <https://www.investopedia.com/terms/m/marketonclose.asp>`__
    MARKET_ON_CLOSE = 'MARKET_ON_CLOSE'

    #: Exercise an option.
    EXERCISE = 'EXERCISE'

    #: Similar to ``STOP_LIMIT``, except if the price moves in your favor, the
    #: stop price is adjusted in that direction. Places a limit order at the
    #: specified price if the stop condition is met.
    #: `More info <https://www.investopedia.com/terms/t/trailingstop.asp>`__.
    TRAILING_STOP_LIMIT = 'TRAILING_STOP_LIMIT'

    #: Place an order for an options spread resulting in a net debit.
    #: `More info <https://www.investopedia.com/ask/answers/042215/
    #: whats-difference-between-credit-spread-and-debt-spread.asp>`__
    NET_DEBIT = 'NET_DEBIT'

    #: Place an order for an options spread resulting in a net credit.
    #: `More info <https://www.investopedia.com/ask/answers/042215/
    #: whats-difference-between-credit-spread-and-debt-spread.asp>`__
    NET_CREDIT = 'NET_CREDIT'

    #: Place an order for an options spread resulting in neither a credit nor a
    #: debit.
    #: `More info <https://www.investopedia.com/ask/answers/042215/
    #: whats-difference-between-credit-spread-and-debt-spread.asp>`__
    NET_ZERO = 'NET_ZERO'

    LIMIT_ON_CLOSE = 'LIMIT_ON_CLOSE'


class ComplexOrderStrategyType(Enum):
    '''
    Explicit order strategies for executing multi-leg options orders.
    '''

    #: No complex order strategy. This is the default.
    NONE = 'NONE'

    #: `Covered call <https://tickertape.tdameritrade.com/trading/
    #: selling-covered-call-options-strategy-income-hedging-15135>`__
    COVERED = 'COVERED'

    #: `Vertical spread <https://tickertape.tdameritrade.com/trading/
    #: vertical-credit-spreads-high-probability-15846>`__
    VERTICAL = 'VERTICAL'

    #: `Ratio backspread <https://tickertape.tdameritrade.com/trading/
    #: pricey-stocks-ratio-spreads-15306>`__
    BACK_RATIO = 'BACK_RATIO'

    #: `Calendar spread <https://tickertape.tdameritrade.com/trading/
    #: calendar-spreads-trading-primer-15095>`__
    CALENDAR = 'CALENDAR'

    #: `Diagonal spread <https://tickertape.tdameritrade.com/trading/
    #: love-your-diagonal-spread-15030>`__
    DIAGONAL = 'DIAGONAL'

    #: `Straddle spread <https://tickertape.tdameritrade.com/trading/
    #: straddle-strangle-option-volatility-16208>`__
    STRADDLE = 'STRADDLE'

    #: `Strandle spread <https://tickertape.tdameritrade.com/trading/
    #: straddle-strangle-option-volatility-16208>`__
    STRANGLE = 'STRANGLE'

    COLLAR_SYNTHETIC = 'COLLAR_SYNTHETIC'

    #: `Butterfly spread <https://tickertape.tdameritrade.com/trading/
    #: butterfly-spread-options-15976>`__
    BUTTERFLY = 'BUTTERFLY'

    #: `Condor spread <https://www.investopedia.com/terms/c/
    #: condorspread.asp>`__
    CONDOR = 'CONDOR'

    #: `Iron condor spread <https://tickertape.tdameritrade.com/trading/
    #: iron-condor-options-spread-your-trading-wings-15948>`__
    IRON_CONDOR = 'IRON_CONDOR'

    #: `Roll a vertical spread <https://tickertape.tdameritrade.com/trading/
    #: exit-winning-losing-trades-16685>`__
    VERTICAL_ROLL = 'VERTICAL_ROLL'

    #: `Collar strategy <https://tickertape.tdameritrade.com/trading/
    #: stock-hedge-options-collars-15529>`__
    COLLAR_WITH_STOCK = 'COLLAR_WITH_STOCK'

    #: `Double diagonal spread <https://optionstradingiq.com/
    #: the-ultimate-guide-to-double-diagonal-spreads/>`__
    DOUBLE_DIAGONAL = 'DOUBLE_DIAGONAL'

    #: `Unbalanced butterfy spread  <https://tickertape.tdameritrade.com/
    #: trading/unbalanced-butterfly-strong-directional-bias-15913>`__
    UNBALANCED_BUTTERFLY = 'UNBALANCED_BUTTERFLY'
    UNBALANCED_CONDOR = 'UNBALANCED_CONDOR'
    UNBALANCED_IRON_CONDOR = 'UNBALANCED_IRON_CONDOR'
    UNBALANCED_VERTICAL_ROLL = 'UNBALANCED_VERTICAL_ROLL'

    #: Mutual fund swap
    MUTUAL_FUND_SWAP = 'MUTUAL_FUND_SWAP'

    #: A custom multi-leg order strategy.
    CUSTOM = 'CUSTOM'


class Destination(Enum):
    '''
    Destinations for when you want to request a specific destination for your
    order.
    '''

    INET = 'INET'
    ECN_ARCA = 'ECN_ARCA'
    CBOE = 'CBOE'
    AMEX = 'AMEX'
    PHLX = 'PHLX'
    ISE = 'ISE'
    BOX = 'BOX'
    NYSE = 'NYSE'
    NASDAQ = 'NASDAQ'
    BATS = 'BATS'
    C2 = 'C2'
    AUTO = 'AUTO'


class StopPriceLinkBasis(Enum):
    MANUAL = 'MANUAL'
    BASE = 'BASE'
    TRIGGER = 'TRIGGER'
    LAST = 'LAST'
    BID = 'BID'
    ASK = 'ASK'
    ASK_BID = 'ASK_BID'
    MARK = 'MARK'
    AVERAGE = 'AVERAGE'


class StopPriceLinkType(Enum):
    VALUE = 'VALUE'
    PERCENT = 'PERCENT'
    TICK = 'TICK'


class StopType(Enum):
    STANDARD = 'STANDARD'
    BID = 'BID'
    ASK = 'ASK'
    LAST = 'LAST'
    MARK = 'MARK'


class PriceLinkBasis(Enum):
    MANUAL = 'MANUAL'
    BASE = 'BASE'
    TRIGGER = 'TRIGGER'
    LAST = 'LAST'
    BID = 'BID'
    ASK = 'ASK'
    ASK_BID = 'ASK_BID'
    MARK = 'MARK'
    AVERAGE = 'AVERAGE'


class PriceLinkType(Enum):
    VALUE = 'VALUE'
    PERCENT = 'PERCENT'
    TICK = 'TICK'


class TaxLotMethod(Enum):
    FIFO = 'FIFO'
    LIFO = 'LIFO'
    HIGH_COST = 'HIGH_COST'
    LOW_COST = 'LOW_COST'
    AVERAGE_COST = 'AVERAGE_COST'
    SPECIFIC_LOT = 'SPECIFIC_LOT'
    LOSS_HARVESTER = 'LOSS_HARVESTER'


class EquityInstruction(Enum):
    '''
    Instructions for opening and closing equity positions.
    '''

    #: Open a long equity position
    BUY = 'BUY'

    #: Close a long equity position
    SELL = 'SELL'

    #: Open a short equity position
    SELL_SHORT = 'SELL_SHORT'

    #: Close a short equity position
    BUY_TO_COVER = 'BUY_TO_COVER'


class OptionInstruction(Enum):
    '''
    Instructions for opening and closing options positions.
    '''
    #: Enter a new long option position
    BUY_TO_OPEN = 'BUY_TO_OPEN'

    #: Exit an existing long option position
    SELL_TO_CLOSE = 'SELL_TO_CLOSE'

    #: Enter a short position in an option
    SELL_TO_OPEN = 'SELL_TO_OPEN'

    #: Exit an existing short position in an option
    BUY_TO_CLOSE = 'BUY_TO_CLOSE'


class SpecialInstruction(Enum):
    '''
    Special instruction for trades.
    '''

    #: Disallow partial order execution.
    #: `More info <https://www.investopedia.com/terms/a/aon.asp>`__.
    ALL_OR_NONE = 'ALL_OR_NONE'

    #: Do not reduce order size in response to cash dividends.
    #: `More info <https://www.investopedia.com/terms/d/dnr.asp>`__.
    DO_NOT_REDUCE = 'DO_NOT_REDUCE'

    #: Combination of ``ALL_OR_NONE`` and ``DO_NOT_REDUCE``.
    ALL_OR_NONE_DO_NOT_REDUCE = 'ALL_OR_NONE_DO_NOT_REDUCE'


class OrderStrategyType(Enum):
    '''
    Rules for composite orders.
    '''

    #: No chaining, only a single order is submitted
    SINGLE = 'SINGLE'

    CANCEL = 'CANCEL'
    RECALL = 'RECALL'
    PAIR = 'PAIR'
    FLATTEN = 'FLATTEN'
    TWO_DAY_SWAP = 'TWO_DAY_SWAP'
    BLAST_ALL = 'BLAST_ALL'

    #: Execution of one order cancels the other
    OCO = 'OCO'

    #: Execution of one order triggers placement of the other
    TRIGGER = 'TRIGGER'


def one_cancels_other(order1, order2):
    '''
    If one of the orders is executed, immediately cancel the other.
    '''
    from schwab.orders.generic import OrderBuilder

    return (OrderBuilder()
            .set_order_strategy_type(OrderStrategyType.OCO)
            .add_child_order_strategy(order1)
            .add_child_order_strategy(order2))


def first_triggers_second(first_order, second_order):
    '''
    If ``first_order`` is executed, immediately place ``second_order``.
    '''
    from schwab.orders.generic import OrderBuilder

    return (first_order
            .set_order_strategy_type(OrderStrategyType.TRIGGER)
            .add_child_order_strategy(second_order))
